# %%
## Import
import os
import numpy as np

import torch
print(torch.__version__)
import torch.nn.functional as F
import torchvision.models
from torchvision import transforms,datasets
from PIL import Image


import matplotlib
import matplotlib.pyplot as plt

#!%load_ext autoreload
#!%autoreload 2
#!%matplotlib inline

## Helper Functions
def show_tensor_image(img_t):
    """
    Function to show the image after initial transform, note: need to permute dimensions from the tensor CxWxH convention
    """
    im = img_t.detach()
    im = torch.squeeze(im, 0) if im.dim()==4 else im
    im = im.permute([1,2,0]) if im.dim()==3 else im
    im = im - im.min()
    im /= im.max()
    return plt.gca().imshow(im.cpu().numpy())

def select_device():
    """ 
    Find the CUDA device with max available memory and set the global dev variable to it
    If less than 4GB memory on all devices, resort to dev='cpu'
    Repeated calls to the function select the same GPU previously selected
    """
    global dev
    global my_gpu
    if 'my_gpu' in globals() and 'cuda' in str(my_gpu):
        dev = my_gpu
    else:
        # find free GPU
        os.system('nvidia-smi -q -d Memory |grep -A4 GPU|grep Used >tmp')
        memory_used = np.array([int(x.split()[2]) for x in open('tmp', 'r').readlines()])
        print(memory_used)
        ii = np.arange(len(memory_used))
        mask = memory_used < 4000
        print(mask)
        if mask.any():
            mask_index = np.argmin(memory_used[mask])
            index = (ii[mask])[mask_index]
            my_gpu = torch.device(index)
        else:
            my_gpu = torch.device('cpu')
        dev = my_gpu


# %%
# Load Network
# model = torchvision.models.vgg11(weights=torchvision.models.VGG11_Weights.DEFAULT)
model = torchvision.models.squeezenet1_0(weights=torchvision.models.SqueezeNet1_0_Weights.DEFAULT)

print(model)
print(model.features[0].weight.device)
dev = model.features[0].weight.device

# %% [markdown]
# ### Test Classification
# 
model.eval()
# Load Test Image
img = Image.open("dog.jpg")
plt.figure()
plt.imshow(img)
plt.show()

# reshape as [C x H x W], normalize
transform = transforms.Compose([transforms.ToTensor(), transforms.Resize(224, antialias=True), transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])])

# reshape as [B C H W], move to device
x = transform(img).unsqueeze(0).to(dev)

# Load labels
with open("imagenet_classes.txt", "r") as f:
    classes = [s.strip() for s in f.readlines()]
    
# TASK PART 1: compute predictions and print top 5 with confidences  
# -------------Implement---------------

# %% [markdown]
# ### Network First Conv Filters and Features
# -------------Implement---------------

# %% [markdown]
# ### Data Loader from Image Folders

# %%
train_data = datasets.ImageFolder('/local/temporary/butterflies/train', transforms.ToTensor())

# %%
# Normalization Statistics
# -------------Implement---------------
mean = 0
std = 0
print(mean)
print(std)

# %%
transform = transforms.Compose([transforms.ToTensor(), transforms.Normalize(mean=mean, std=std)])
train_data = datasets.ImageFolder('/local/temporary/butterflies/train', transform)
# split train data into training and validation
# -------------Implement---------------
# test set, using the same transform
test_data = datasets.ImageFolder('/local/temporary/butterflies/test', transform)

# %% [markdown]
# ### Finetune last layer

# %%
# Move model to device
select_device()
model.to(dev)
print(model.features[0].weight.device)

# %%
def evaluate(model, loader):
    """ Evaluate the model with the given dataset loader """
    #---------Implement-----------
    pass

# %%
epochs = 10
model.eval()
# freeze all layers but the last one, re-initialze last layer
# ----------Implement----------

# optimizer
optimizer = torch.optim.SGD(model.parameters(), lr = 0.01)
# train and validation loaders
train_loader = torch.utils.data.DataLoader(train_set, batch_size=8, shuffle=True, num_workers=0)
val_loader = torch.utils.data.DataLoader(val_set, batch_size=8, shuffle=True, num_workers=0)
# trianing loop
for e in range(epochs):
    for (x,t) in train_loader:
        x = x.to(dev)
        t = t.to(dev)
        score = model(x)
        log_p = torch.log_softmax(score, -1)
        l = F.nll_loss(log_p, t, reduction='sum')
        optimizer.zero_grad()
        l.backward()
        optimizer.step()
    # ----------Implement----------
    # Compute loss and accuracy using training set
    # Compute loss and accuracy using validation set
    # Save so-far best model and its validation accuracy

# %%
## Test
test_loader = torch.utils.data.DataLoader(test_data, batch_size=8, shuffle=False, num_workers=0)
evaluate(test_loader)
# Error case analyzis
# ----------Implement----------

# %% [markdown]
# ### Part 5: Data Augmentation

